from pandas_ml_quant_data_provider.datafetching import fetch_yahoo, fetch_investing, fetch_cryptocompare_daily, \
    fetch_cryptocompare_hourly, fetch_fred

YAHOO = fetch_yahoo
INVESTING = fetch_investing
CRYPTO_COMPARE = fetch_cryptocompare_daily
CRYPTO_COMPARE_HOURLY = fetch_cryptocompare_hourly
FRED = fetch_fred

PROVIDER_MAP = {
    "yahoo": YAHOO,
    "investing": INVESTING,
    "cryptocompare": CRYPTO_COMPARE,
    "cryptocompare_daily": CRYPTO_COMPARE,
    "cryptocompare_hourly": CRYPTO_COMPARE_HOURLY,
    "fred": FRED
}


import os
from unittest import TestCase

from pandas_ml_quant_data_provider import fetch_timeseries, YAHOO, INVESTING, CRYPTO_COMPARE, FRED, pd


class TestFetch(TestCase):

    def test__mixed(self):
        df = fetch_timeseries({
            YAHOO: ["SPY", "DIA"],
            INVESTING: ["index::NYSE Tick Index::united states", "bond::U.S. 30Y::united states"],
            CRYPTO_COMPARE: ["BTC"],
        })

        print(df.columns.to_list())
        print(df.head())
        print(len(df))

        self.assertListEqual(
            [('SPY', 'Open'), ('SPY', 'High'), ('SPY', 'Low'), ('SPY', 'Close'), ('SPY', 'Volume'), ('SPY', 'Dividends'), ('SPY', 'Stock Splits'),
             ('DIA', 'Open'), ('DIA', 'High'), ('DIA', 'Low'), ('DIA', 'Close'), ('DIA', 'Volume'), ('DIA', 'Dividends'), ('DIA', 'Stock Splits'),
             ('NYSE Tick Index', 'Open'), ('NYSE Tick Index', 'High'), ('NYSE Tick Index', 'Low'), ('NYSE Tick Index', 'Close'), ('NYSE Tick Index', 'Volume'), ('NYSE Tick Index', 'Currency'),
             ('U.S. 30Y', 'Open'), ('U.S. 30Y', 'High'), ('U.S. 30Y', 'Low'), ('U.S. 30Y', 'Close'),
             ('BTC', 'close'), ('BTC', 'high'), ('BTC', 'low'), ('BTC', 'open'), ('BTC', 'volumefrom'), ('BTC', 'volumeto')],
            df.columns.to_list()
        )

        self.assertGreaterEqual(len(df), 2400)

    def test__fred(self):

        if "FRED_API_KEY" not in os.environ:
            print("skipping FRED due to missing FRED_API_KEY key")
            return

        df = fetch_timeseries({FRED: ["GDP", "WALCL"]}, ffill=True)

        print(df.columns.to_list())
        print(df.head())
        print(len(df))

        self.assertListEqual(df.columns.to_list(), [('FRED', 'GDP'), ('FRED', 'WALCL')])
        self.assertGreaterEqual(len(df), 968)

    def test__cryptocompare(self):
        df = fetch_timeseries({"cryptocompare_hourly": "BTC"}, force_lower_case=True)
        print(df.tail())

        self.assertGreaterEqual(len(df), 1440)

    def test__monkey_patch(self):
        self.assertIsNotNone(pd.fetch_timeseries)
        self.assertIsNotNone(pd.fetch_yahoo)
        self.assertIsNotNone(pd.fetch_investing)
        self.assertIsNotNone(pd.fetch_cryptocompare_daily)




import logging

import cachetools.func
import cachetools.func
import investpy

from pandas_ml_common.utils import inner_join, add_multi_index

_log = logging.getLogger(__name__)

START_DATE = '01/01/1960'
STOP_DATE = '31/12/2029'

ASSET_CATEGORY = {
    "bond": lambda s, c: investpy.get_bond_historical_data(bond=s, from_date=START_DATE, to_date=STOP_DATE),
    "certificate": lambda s, c: investpy.get_certificate_historical_data(certificate=s, country=c, from_date=START_DATE, to_date=STOP_DATE),
    "commodity": lambda s, c: investpy.get_commodity_historical_data(commodity=s, from_date=START_DATE, to_date=STOP_DATE),
    "crypto": lambda s, c: investpy.get_crypto_historical_data(crypto=s, from_date=START_DATE, to_date=STOP_DATE),
    "currency": lambda s, c: investpy.get_currency_cross_historical_data(currency_cross=s, from_date=START_DATE, to_date=STOP_DATE),
    "etf": lambda s, c: investpy.get_etf_historical_data(etf=s, country=c, from_date=START_DATE, to_date=STOP_DATE),
    "fund": lambda s, c: investpy.get_fund_historical_data(fund=s, country=c, from_date=START_DATE, to_date=STOP_DATE),
    "index": lambda s, c: investpy.get_index_historical_data(index=s, country=c, from_date=START_DATE, to_date=STOP_DATE),
    "stock": lambda s, c: investpy.get_stock_historical_data(stock=s, country=c, from_date=START_DATE, to_date=STOP_DATE),
}


@cachetools.func.ttl_cache(maxsize=1, ttl=10 * 60)
def fetch_investing(*args: str, multi_index: bool = False, **kwargs):
    df = None

    for compound_symbol in args:
        details = compound_symbol.split("::")
        asset_class, symbol, country = details if len(details) >= 3 else (*details, None)
        _df = _fetch_investing(asset_class, symbol, country)

        if multi_index:
            _df = add_multi_index(_df, symbol, True)

        if df is None:
            df = _df.add_prefix(symbol) if len(args) > 1 and not multi_index else _df
        else:
            df = inner_join(df, _df, prefix='' if multi_index else symbol, force_multi_index=multi_index)

    return df


@cachetools.func.ttl_cache(maxsize=1, ttl=10 * 60)
def _fetch_investing(asset_category, symbol, country):
    try:
        return ASSET_CATEGORY[asset_category](symbol, country)
    except IndexError as ie:
        _log.warning(f"{ie}, {asset_category}, {symbol}, {country}")




from .fetch_cryptocompare import fetch_cryptocompare_daily, fetch_cryptocompare_hourly
from .fetch_yahoo import fetch_yahoo
from .read_ts_file import read_ts_csv
from .fetch_investing import fetch_investing
from .fetch_fred import fetch_fred


from typing import List, Callable, Dict, Any, Tuple, Set

import pandas as pd

from pandas_ml_common.utils import call_callable_dynamic_args, add_multi_index, inner_join
from pandas_ml_quant_data_provider.provider import PROVIDER_MAP


def fetch_timeseries(providers: Dict[Callable[[Any], pd.DataFrame], List[str]],
                     start_date: str = None,
                     force_lower_case: bool = False,
                     multi_index: bool = None,
                     ffill: bool = False,
                     **kwargs):
    symbol_type = (List, Tuple, Set)
    expected_frames = sum(len(s) if isinstance(s, symbol_type) else 1 for s in providers.values())
    df = None

    if multi_index is None and expected_frames > 1:
        multi_index = True

    for provider, symbols in providers.items():
        # make sure provider is an actual provider -> a callable
        if not callable(provider):
            provider = PROVIDER_MAP[provider]

        # make sure the symbols are iterable -> wrap single symbols into a list
        if not isinstance(symbols, symbol_type):
            symbols = [symbols]

        # fetch all symbols of all providers (later we could do this in parallel)
        for symbol in symbols:
            _df = call_callable_dynamic_args(provider, symbol, multi_index=multi_index, **kwargs)

            if _df is None:
                continue

            if multi_index:
                if not isinstance(_df.columns, pd.MultiIndex):
                    _df = add_multi_index(_df, symbol, True)

                if force_lower_case:
                    _df.columns = pd.MultiIndex.from_tuples([(h.lower(), c.lower()) for h, c in _df.columns.to_list()])
            else:
                if isinstance(_df.columns, pd.MultiIndex):
                    _df.columns = [t[-1]for t in _df.columns.to_list()]

                if force_lower_case:
                    _df.columns = [c.lower() for c in _df.columns.to_list()]

            if df is None:
                df = _df
            else:
                df = inner_join(df, _df, force_multi_index=multi_index, ffill=ffill)

    return df if start_date is None else df[start_date:]




import logging
import unittest

from pandas_ml_quant_data_provider import pd

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class ComponentTest(unittest.TestCase):

    def test_fetch_yahoo(self):
        """when"""
        df = pd.fetch_yahoo(spy="SPY").tail()
        print(df.columns)

        """then"""
        self.assertTrue(df["spy_Close"].sum() > 0)


    def test_crypto_compare(self):
        """when"""
        df = pd.fetch_cryptocompare_daily(coin="BTC", limit=None).tail()
        print(df.columns)

        """then"""
        self.assertTrue(df["close"].sum() > 0)




import logging
import traceback

import cachetools.func

from pandas_ml_common import pd
from pandas_ml_common.utils import merge_kwargs, inner_join


@cachetools.func.ttl_cache(maxsize=1, ttl=10 * 60)
def fetch_yahoo(*args: str, period: str = 'max', multi_index: bool = False, **kwargs: str):
    df = None

    if len(args) == 1:
        df = __download_yahoo_data(args[0], period)
    else:
        # convert args to kwargs
        if len(args) > 0:
            kwargs = merge_kwargs({arg: arg for arg in args}, kwargs)

        for k, v in kwargs.items():
            px = f'{k}_'
            df_ = __download_yahoo_data(v, period)

            if multi_index:
                df_.columns = pd.MultiIndex.from_product([[k], df_.columns])

                if df is None:
                    df = df_
                else:
                    df = inner_join(df, df_)
            else:
                if df is None:
                    df = df_.add_prefix(px)
                else:
                    df = inner_join(df, df_, prefix=px)

    # print some statistics
    if df is None:
        logging.warning("nothing downloaded")
    else:
        logging.info(f'number of rows for joined dataframe = {len(df)}, from {df.index[0]} to {df.index[-1]}')

    return df


def __download_yahoo_data(symbol, period):
    import yfinance as yf
    df = None

    # bloody skew index does not have any data on yahoo
    if symbol == '^SKEW':
        df = pd.read_csv('http://www.cboe.com/publish/scheduledtask/mktdata/datahouse/skewdailyprices.csv',
                          skiprows=1,
                          parse_dates=True,
                          index_col='Date') \
            .drop(['Unnamed: 2', 'Unnamed: 3'], axis=1)
    else:
        ticker = yf.Ticker(symbol)
        try:
            # first try to append the most recent data
            df = ticker.history(period="1d", interval='1d')[-1:].combine_first(ticker.history(period=period))
        except IOError:
            traceback.print_exc()
            logging.warning(
                'failed to add yf.Ticker({v}).history(period="1d", interval="1d")[-1:] fallback to hist only!')
            df = ticker.history(period=period)

    # print some statistics
    logging.info(f'number of rows for {symbol} = {len(df)}, from {df.index[0]} to {df.index[-1]} period={period}')

    return df



import datetime

import cachetools
import cachetools.func
import pandas as pd
import pytz

from .extern.cryptocompare import CURR, LIMIT, TIME, get_historical_price_day, get_historical_price_hour


@cachetools.func.ttl_cache(maxsize=1, ttl=10 * 60)
def fetch_cryptocompare_daily(coin, curr=CURR, limit=None, **kwargs) -> pd.DataFrame:
    data = get_historical_price_day(coin, curr, limit)
    return _data_to_frame(data["Data"])


@cachetools.func.ttl_cache(maxsize=1, ttl=10 * 60)
def fetch_cryptocompare_hourly(coin, curr=CURR, limit=LIMIT, **kwargs) -> pd.DataFrame:
    data = get_historical_price_hour(coin, curr, limit)
    return _data_to_frame(data["Data"])


def _data_to_frame(data):
    df = pd.DataFrame(data)
    df.index = pd.DatetimeIndex(
        df[TIME].apply(lambda t: datetime.datetime.fromtimestamp(t, tz=pytz.utc).replace(tzinfo=None)))

    return df.drop(TIME, axis=1).sort_index(axis=0)


# slightly enhanced version of https://github.com/lagerfeuer/cryptocompare/blob/f940fab908a9b79ce9069cc6dea9cc5a3f2e2eee/cryptocompare/cryptocompare.py

import datetime
import logging
import time

import cachetools.func
import requests

_log = logging.getLogger(__name__)


# API
URL_COIN_LIST = 'https://www.cryptocompare.com/api/data/coinlist/'
URL_PRICE = 'https://min-api.cryptocompare.com/data/pricemulti?fsyms={}&tsyms={}'
URL_PRICE_MULTI = 'https://min-api.cryptocompare.com/data/pricemulti?fsyms={}&tsyms={}'
URL_PRICE_MULTI_FULL = 'https://min-api.cryptocompare.com/data/pricemultifull?fsyms={}&tsyms={}'
URL_HIST_PRICE = 'https://min-api.cryptocompare.com/data/pricehistorical?fsym={}&tsyms={}&ts={}&e={}'
URL_HIST_PRICE_DAY = 'https://min-api.cryptocompare.com/data/histoday?fsym={}&tsym={}&limit={}&allData={}'
URL_HIST_PRICE_HOUR = 'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym={}&limit={}&toTs={}'
URL_HIST_PRICE_MINUTE = 'https://min-api.cryptocompare.com/data/histominute?fsym={}&tsym={}&limit={}'
URL_AVG = 'https://min-api.cryptocompare.com/data/generateAvg?fsym={}&tsym={}&e={}'
URL_EXCHANGES = 'https://www.cryptocompare.com/api/data/exchanges'

# MAX
MAX_LIMIT = 2000
MAX_INT = 2**31 - 1

# FIELDS
PRICE = 'PRICE'
HIGH = 'HIGH24HOUR'
LOW = 'LOW24HOUR'
VOLUME = 'VOLUME24HOUR'
CHANGE = 'CHANGE24HOUR'
CHANGE_PERCENT = 'CHANGEPCT24HOUR'
MARKETCAP = 'MKTCAP'
DATA = 'Data'
TIME = 'time'

# DEFAULTS
CURR = 'USD'
LIMIT = 1440
###############################################################################


@cachetools.func.ttl_cache(maxsize=1, ttl=10 * 60)
def query_cryptocompare(url,errorCheck=True):
    try:
        _log.debug(url)
        response = requests.get(url).json()
    except Exception as e:
        _log.error('Error getting coin information. %s' % str(e))
        return None
    if errorCheck and (response.get('Response') == 'Error'):
        _log.error('[ERROR] %s' % response.get('Message'))
        return None
    return response


def format_parameter(parameter):
    if isinstance(parameter, list):
        return ','.join(parameter)
    else:
        return parameter

###############################################################################


def get_coin_list(format=False):
    response = query_cryptocompare(URL_COIN_LIST, False)['Data']
    if format:
        return list(response.keys())
    else:
        return response


# TODO: add option to filter json response according to a list of fields
def get_price(coin, curr=CURR, full=False):
    if full:
        return query_cryptocompare(URL_PRICE_MULTI_FULL.format(format_parameter(coin),
            format_parameter(curr)))
    if isinstance(coin, list):
        return query_cryptocompare(URL_PRICE_MULTI.format(format_parameter(coin),
            format_parameter(curr)))
    else:
        return query_cryptocompare(URL_PRICE.format(coin, format_parameter(curr)))


def get_historical_price(coin, curr=CURR, timestamp=time.time(), exchange='CCCAGG'):
    if isinstance(timestamp, datetime.datetime):
        timestamp = time.mktime(timestamp.timetuple())
    return query_cryptocompare(URL_HIST_PRICE.format(coin, format_parameter(curr),
        int(timestamp), format_parameter(exchange)))


def get_historical_price_day(coin, curr=CURR, limit=LIMIT):
    all_data = "false"

    if limit is None or limit > LIMIT:
        limit = 1
        all_data = "true"

    return query_cryptocompare(URL_HIST_PRICE_DAY.format(coin, format_parameter(curr), limit, all_data))


def get_historical_price_hour(coin, curr=CURR, limit=LIMIT):
    current_ts = int(time.time()) + 60 * 60 * 24 + 1 # to be on the safe side
    if limit is None or limit > LIMIT:
        _log.info("batch download < now")
        youngest_ts = get_historical_price_day(coin, curr, None)["TimeFrom"] / 1000
        data = query_cryptocompare(URL_HIST_PRICE_HOUR.format(coin, format_parameter(curr), MAX_LIMIT, current_ts))
        batch = data

        while True:
            last_ts = min(batch[DATA][0][TIME], batch[DATA][-1][TIME])
            _log.info(f"batch download < {last_ts}")
            batch = query_cryptocompare(URL_HIST_PRICE_HOUR.format(coin, format_parameter(curr), MAX_LIMIT, last_ts - 1))
            if batch is None:
                return data
            else:
                data[DATA] += batch[DATA]
                if len(batch) <= 0 or last_ts < youngest_ts:
                    return data
                elif limit is not None and len(data[DATA]) >= limit:
                    data[DATA] = data[DATA][:limit]
                    return data
    else:
        return query_cryptocompare(URL_HIST_PRICE_HOUR.format(coin, format_parameter(curr), limit, current_ts))


def get_historical_price_minute(coin, curr=CURR, limit=LIMIT):
    return query_cryptocompare(URL_HIST_PRICE_MINUTE.format(coin, format_parameter(curr), limit))


def get_avg(coin, curr=CURR, exchange='CCCAGG'):
    response = query_cryptocompare(URL_AVG.format(coin, curr, format_parameter(exchange)))
    if response:
        return response['RAW']


def get_exchanges():
    response = query_cryptocompare(URL_EXCHANGES)
    if response:
        return response['Data']



"""Augment pandas DataFrame with methods to fetch time series data for quant finance"""
__version__ = '0.1.9'

import pandas as pd

import pandas_ml_quant_data_provider.datafetching as data_fetchers
from .fetch import fetch_timeseries
from .provider import *

setattr(pd, "fetch_timeseries", fetch_timeseries)
setattr(pd, "read_ts_csv", data_fetchers.read_ts_csv)

# add data fetcher functions
for fetcher_functions in [data_fetchers]:
    for fetcher_function in dir(fetcher_functions):
        if fetcher_function.startswith("fetch_"):
            setattr(pd, fetcher_function, getattr(fetcher_functions, fetcher_function))



import logging
import os
from typing import List, Tuple, Iterator

import cachetools.func
import cachetools.func
from fredapi import Fred

from pandas_ml_common.utils import add_multi_index

_log = logging.getLogger(__name__)


@cachetools.func.ttl_cache(maxsize=1, ttl=10 * 60)
def fetch_fred(symbols, multi_index=False):
    if "FRED_API_KEY" not in os.environ:
        raise ValueError("you need to set the environment variable `FRED_API_KEY`. You can get a key as described here:"
                         "https://github.com/mortada/fredapi")

    fred = Fred()
    df = None

    if not isinstance(symbols, (List, Tuple, Iterator)):
        symbols = [symbols]

    for symbol in symbols:
        series = fred.get_series(symbol)

        if not series.name:
            series.name = symbol

        if df is None:
            df = series.to_frame()
        else:
            df = df.join(series)

    return add_multi_index(df, "FRED", True) if multi_index else df